/*
 * PHexGrid.java
 *
 * 2d hexagonal grid of crypts.
 * see Clark Verbrugge's Hex Grids article for details (http://www.sable.mcgill.ca/~clump/hexes.txt):
                Hex X coord 

    \   \   \   \   \   \   \   \   \  
     \   \   \   \   \   \   \   \   \  
      0   1   2   3   4   5   6   7   8     
       / \ / \ / \ / \ / \ / \ / \ / \ / \   
      | O | W |   | A |   |   |   |   |   |  
H      \ / \ / \ / \ / \ / \ / \ / \ / \ / \ 
e   0   | V |   | Z |   |   |   |   |   |   |
x  /   / \ / \ / \ / \ / \ / \ / \ / \ / \ / 
  /   | U |   |   |   |   |   |   |   |   |  
Y      \ / \ / \ / \ / \ / \ / \ / \ / \ / \ 
    1   | B |   |   |   |   |   |   |   |   |
c  /   / \ / \ / \ / \ / \ / \ / \ / \ / \ / 
o /   |   |   |   |   |   |   |   |   |   |  
o      \ / \ / \ / \ / \ / \ / \ / \ / \ / \ 
r   2   |   |   |   |   |   |   |   |   |   |
d  /   / \ / \ / \ / \ / \ / \ / \ / \ / \ / 
  /   |   |   |   |   |   |   |   |   |   |  
       \ / \ / \ / \ / \ / \ / \ / \ / \ / \ 
    3   |   |   |   |   |   |   |   |   |   |
   /     \ / \ / \ / \ / \ / \ / \ / \ / \ / 
  /

 in array space,
 U=(0,2), V=(0,1), W=(1,0), Z=(2,1), A=(3,0), B=(0,3)
*/

package sim;

import java.util.*;

public class PHexGrid extends PGrid {
    // Class constants
    public final static double HEXTISSUEVERTICALSCALINGFACTOR = 0.75; // (1+Math.sin(Math.toRadians(30)))/2.0
	
    // Data Members
    private int _nXSize,_nYSize;  // dimensions of hex grid in crypts
    private Boolean _bWrapX;  // cyclic boundary on X axis
    private Boolean _bWrapY;  // cyclic boundary on Y axis
	
    // Constructor - allows caller to specify whether the vertical and horizontal edges wrap aroun
    public PHexGrid(int nXSize, int nYSize, CryptFactory factory, TissueEnvironmentSettings params, boolean bWrapX, boolean bWrapY) {
	super();
	_factory = factory;

	if (!bWrapX && bWrapY) {  // if boundaries must wrap
	    System.err.println("Error in PHexGrid:PHexGrid: If bWrapY is true, then bWrapX must be true.");
	    System.exit(-1);
	}
	_nXSize = nXSize;
	_nYSize = nYSize;
	_bWrapX = bWrapX;
	_bWrapY = bWrapY;
	_crypts = new Crypt[_nXSize*_nYSize];
	for (int i=0; i<_crypts.length; i++) {
	    _crypts[i] = factory.createCrypt();
	    _crypts[i].setGrid(this);
	    _crypts[i].setIndex(i);
	    _crypts[i].setParams(params);
	}
	for (int i=0; i<_crypts.length; i++) {
	    ArrayList n = getNeighbors(i);
	    for (int k=0; k<n.size(); k++) {
		Integer c = (Integer)n.get(k);
		_crypts[i].addNeighbor(_crypts[c.intValue()]);
	    }
	}
    }

    // Constructor - creates tissue with hard (non-cyclic) boundaries
    public PHexGrid(int nXSize, int nYSize, CryptFactory factory, TissueEnvironmentSettings params) {
	this(nXSize, nYSize, factory, params, false, false);
    }

    // Copy Constructor
    public PHexGrid( PHexGrid rhs ) {

	// Don't copy the base class's members, rather just init them
	super();
		
	// JTE TODO -- throw exception if rhs._factory is null
	_factory = rhs._factory;
		
	this._nXSize = rhs._nXSize;
	this._nYSize = rhs._nYSize;
	this._bWrapX = rhs._bWrapX;
	this._bWrapY = rhs._bWrapY;
	
	// Copy the crypts element by element (we need to do a deep copy, so arraycopy() won't suffice)
	_crypts = new Crypt[rhs._crypts.length];
	for(int i=0; i<_crypts.length; i++ ) {
			
	    _crypts[i] = _factory.createCrypt();
	    _crypts[i]._nVertexIndex = rhs._crypts[i]._nVertexIndex;
	    _crypts[i]._grid = rhs._crypts[i]._grid;
			
	    // JTE TODO -- this is really bad casting !!			
	    _crypts[i].SetCryptState( rhs._crypts[i].GetCryptState()  );
	}
		
	// Add each crypt's neighbors
	for (int i=0; i<_crypts.length; i++) {
	    _crypts[i]._neighbors = new ArrayList<Crypt>();		
			
	    ArrayList n = getNeighbors(i);
	    for (int k=0; k<n.size(); k++) {
		Integer c = (Integer)n.get(k);
		_crypts[i].addNeighbor(_crypts[c.intValue()]);
	    }
	}			
    }
	
    /* getNeighbors
     * returns coordinates of all crypts touching crypt n
     */
    public ArrayList<Integer> getNeighbors(int n) {
	ArrayList<Integer> neighbors = new ArrayList<Integer>();
	int x = n%_nXSize;
	int y = n/_nXSize;
	if (!_bWrapX && !_bWrapY) {
	    if (y>0) {
		neighbors.add(new Integer(x + (y-1)*_nXSize));
		if (y%2==0) {
		    if (x>0)
                        neighbors.add(new Integer((x-1) + (y-1)*_nXSize));
		} else if (y%2==1 && x+1<_nXSize) 
		    neighbors.add(new Integer((x+1) + (y-1)*_nXSize));
	    }
	    if (x>0)
		neighbors.add(new Integer((x-1) + y*_nXSize));
	    if (x+1<_nXSize)
		neighbors.add(new Integer((x+1) + y*_nXSize));
	    if (y+1<_nYSize) {
		neighbors.add(new Integer(x + (y+1)*_nXSize));
		if (y%2==0) {
		    if (x>0)
                        neighbors.add(new Integer((x-1) + (y+1)*_nXSize));
		} else if (y%2==1 && x+1<_nXSize)
		    neighbors.add(new Integer((x+1) + (y+1)*_nXSize));
	    }
	} else if (_bWrapX && !_bWrapY) {
	    int nextx = (x+1)%_nXSize;
	    int prevx = ((x-1)+_nXSize)%_nXSize;
	    int nexty = (y+1)%_nYSize;
	    int prevy = ((y-1)+_nYSize)%_nYSize;
	    if (y>0) {
		neighbors.add(new Integer(x + prevy*_nXSize));
		if (y%2==0) {
		    neighbors.add(new Integer(prevx + prevy*_nXSize));
		} else if (y%2==1)
		    neighbors.add(new Integer(nextx + prevy*_nXSize));
	    }
	    neighbors.add(new Integer(prevx + y*_nXSize));
	    neighbors.add(new Integer(nextx + y*_nXSize));
	    if (nexty<_nYSize) {
		neighbors.add(new Integer(x + nexty*_nXSize));
		if (y%2==0) {
		    neighbors.add(new Integer(prevx + nexty*_nXSize));
		} else if (y%2==1)
		    neighbors.add(new Integer(nextx + nexty*_nXSize));
	    }
	} else if (_bWrapX && _bWrapY) {
	    int nextx = (x+1)%_nXSize;
	    int prevx = ((x-1)+_nXSize)%_nXSize;
	    int nexty = (y+1)%_nYSize;
	    int prevy = ((y-1)+_nYSize)%_nYSize;
	    neighbors.add(new Integer(x + prevy*_nXSize));
	    if (y%2==0) {
		neighbors.add(new Integer(prevx + prevy*_nXSize));
	    } else if (y%2==1)
		neighbors.add(new Integer(nextx + prevy*_nXSize));
	    neighbors.add(new Integer(prevx + y*_nXSize));
	    neighbors.add(new Integer(nextx + y*_nXSize));
	    neighbors.add(new Integer(x + nexty*_nXSize));
	    if (y%2==0) {
		neighbors.add(new Integer(prevx + nexty*_nXSize));
	    } else if (y%2==1)
		neighbors.add(new Integer(nextx + nexty*_nXSize));
	}
	return neighbors;
    }
	
    public double getMaxX() {
	return (_nXSize*2.0 + 1.0)/2.0;
    }
	
    public double getMaxY() {
	return _nYSize*HEXTISSUEVERTICALSCALINGFACTOR;
    }
	
    public int getWidth() {
	return _nXSize;
    }
	
    public int getHeight() {
	return _nYSize;
    }
	
    public double getX(int n) {
	int x = n%_nXSize;
	int y = n/_nXSize;
	//		if (x+(y/2) >= _nXSize)
	//			x -= _nXSize;
	return x+((y%2)/2.0);
    }
	
    public double getY(int n) {
	int x = n%_nXSize;
	int y = n/_nXSize;
	//		if (x+(y/2) >= _nXSize)
	//			x -= _nXSize;
	return y*HEXTISSUEVERTICALSCALINGFACTOR;
    }

    public Crypt getCryptAtCenter() {
	return getCrypt((_nYSize/2)*_nXSize + _nXSize/2);
    }
}
