/*
 * SimulationEvent
 */

package sim;

import java.util.*;

public class SimulationEvent implements Comparable<SimulationEvent> {

	// enum type
	public enum EventType { death, mutation, division, 
				wound,                    // wounding event
				seedMutant,               // plant one mutant crypt in middle of tissue
				callback,                 // callback for control loop
                                endSimulation};           // end the simulation
	
	// Data members
	private Crypt _cryptRef; 	// the object to whom this event pertains
	private double _fTime;				// the time in days at which this event will occur
	private EventType _eventType;  			// the event type
    private double _fParam;                         // some events require a parameter
                                                    // for wounding, it is the radius
	private boolean _isValid;				// is this event valid for processing?
	
	// Constructors
	public SimulationEvent(Crypt ref, double time, double param, EventType type) 
	{
		_cryptRef = ref;
		_fTime = time;
		_eventType = type;
		_fParam = param;
		_isValid = true;
	}

        // Constructor that sets "param" to 0.0
	public SimulationEvent(Crypt ref, double time, EventType type) 
	{
		_cryptRef = ref;
		_fTime = time;
		_eventType = type;
		_fParam = 0.0;
		_isValid = true;
	}

	// Copy ctr
	public SimulationEvent( SimulationEvent rhs )
		throws Exception
	{
		copyFrom(rhs);
	}
	
	public SimulationEvent copyFrom( SimulationEvent rhs)
		throws Exception
	{
		if( rhs == null )
			throw new Exception( "Cannot copy null reference of SimulationEvent");
		
		copy( this, rhs );
		return this;
	}
	
	private void copy( SimulationEvent lhs, SimulationEvent rhs)
	{
		lhs._cryptRef = rhs._cryptRef;
		lhs._fTime = rhs._fTime;
		lhs._eventType = rhs._eventType;
		lhs._fParam = rhs._fParam;
		lhs._isValid = rhs._isValid;
	}
	
	// Accessors/mutators
	public EventType get_eventType() {
		return _eventType;
	}

	public void set_eventType(EventType type) {
		_eventType = type;
	}

	public double get_time() {
		return _fTime;
	}

	public void set_time(double time) {
		_fTime = time;
	}

	public double get_param() {
		return _fParam;
	}

	public void set_param(double param) {
		_fParam = param;
	}

	public Crypt get_cryptRef() {
		return _cryptRef;
	}

	public void set_cryptRef(Crypt ref) {
		_cryptRef = ref;
	}

	public boolean isValid() {
		return _isValid;
	}

	public void set_isValid(boolean valid) {
		_isValid = valid;
	}
	
	public void Invalidate() {
		_isValid = false;
	}
	

	
	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 * 
	 * Equality based on hashcode
	 */
    //	@Override
    //	public boolean equals(Object obj) {
    //		return ( this.hashCode() == obj.hashCode() );
		// NOT SURE IF THIS WORKS - DLC
    //	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
    //    	@Override
    //	public int hashCode() {
	    // TODO Auto-generated method stub
    //	    return super.hashCode();
    //    	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {

		String sInfo = "[Simulation Event: type= " + _eventType + "; time = " + _fTime + "; isValid = " + _isValid + "; crypt = " + _cryptRef + "]";
		return sInfo;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(T)
	 * 
	 * SimulationEvents are ordered first by timestamp; 
	 * 
	 * (OLD: then to break ties if the objects are the same,
	 * event types are ordered: death, mutation, division)
	 * 
	 */
	// JTE TODO This must be highly optimized as it will be called excessively!!!!!
	public int compareTo(SimulationEvent that)
	{
		// Unequal event times are easy...
		if( this._fTime != that._fTime ) {
			return ( ( this._fTime < that._fTime ? -1 : 1 ) );
		}
		
		// Otherwise just order by the event type
		if (this._eventType != that._eventType)
		    return this._eventType.compareTo(that._eventType);

		// otherwise order by crypt number
		if (this._cryptRef!=null && that._cryptRef!=null) {
		    if (this._cryptRef.getIndex() < that._cryptRef.getIndex())
			return -1;
		    if (this._cryptRef.getIndex() > that._cryptRef.getIndex())
			return 1;
		}

		// Check for identity
		if( this.equals(that))
		    return 0;
		
		// we should never get here!!!!!!!
		System.out.println(this);
		System.out.println(that);
		assert false;
		System.exit(-1);
		return -1;
	}	
	
	public static class SimulationEventComparator implements Comparator<SimulationEvent> {

		/* (non-Javadoc)
		 * @see java.util.Comparator#compare(T, T)
		 */
		public int compare(SimulationEvent o1, SimulationEvent o2) {
			return o1.compareTo( o2 );
		}
		
	}
}
