/*
 * PNASDriver
 * Batchmode driver used for the PNAS 2008 paper, "Pre-neoplastic 
 * lesion growth driven by the death of adjacent normal stem cells"
 * Input the "experiment number" (from 1 to 23) on the command line, 
 * and the program outputs lesion sizes over time to standard out
 * and sometimes postscript images of simulated lesions.
 */

package ui;

import java.awt.*;
import java.awt.image.*;
import java.io.*;
import javax.imageio.*;
import java.util.ArrayList;
import java.awt.print.*;
import javax.print.*;
import javax.print.attribute.*;
import javax.print.attribute.standard.*;
import java.awt.geom.AffineTransform;

import sim.*;
import lib.KnuthRandom;

public class PNASDriver 
             implements ISimulationProgressCallback, Printable {
    private Scheduler _scheduler;
    private PGrid _grid;
    private PGridView _gridView;
    private int _nTrial;
    private TissueEnvironmentSettings _params;

    public PNASDriver(TissueEnvironmentSettings params) {
	_params = params;
	try
	    {
		_grid = new PHexGrid(params.get_gridSizeX(), params.get_gridSizeY(), CryptFactory.getFactory(), params);
		_scheduler = new Scheduler( _grid, -1, params );
		_scheduler.PostEvent(new SimulationEvent(null, 0.0, SimulationEvent.EventType.seedMutant)); // always put one mutant crypt in the middle
		if (params.get_woundingMode() == TissueEnvironmentSettings.WoundingMode.Distributed && params.get_woundIntervalTime()>100) { // this is a one-time wounding event at day 0
		    _scheduler.PostEvent(new SimulationEvent(null, 0.0, 0, SimulationEvent.EventType.wound ));
		}
	    }
	catch( Exception ex )
	    {
		System.err.println( "Exception caught in GUITissueDriver:GuiTissueDriver(): " + ex.getMessage());
		ex.printStackTrace(System.err);
	    }

    }
    
    // alternate constructor that allows mutant to occur at any time
    // seedtime - what day to create the random mutant
    public PNASDriver(TissueEnvironmentSettings params, double seedtime) {
	_params = params;
	try
	    {
		_grid = new PHexGrid(params.get_gridSizeX(), params.get_gridSizeY(), CryptFactory.getFactory(), params);
		_scheduler = new Scheduler( _grid, -1, params );
		_scheduler.PostEvent(new SimulationEvent(null, seedtime, SimulationEvent.EventType.seedMutant)); // always put one mutant crypt in the middle
		if (params.get_woundingMode() == TissueEnvironmentSettings.WoundingMode.Distributed && params.get_woundIntervalTime()>100) { // this is a one-time wounding event at day 0
		    _scheduler.PostEvent(new SimulationEvent(null, 0.0, 0, SimulationEvent.EventType.wound ));
		}
	    }
	catch( Exception ex )
	    {
		System.err.println( "Exception caught in GUITissueDriver:GuiTissueDriver(): " + ex.getMessage());
		ex.printStackTrace(System.err);
	    }

    }
    
    public long getRandomSeed() {
	return _scheduler.GetRNG().getSeed();
    }

    public void Run() {
	Scheduler.SimulationTerminationReason eTermReason = _scheduler.RunSimulation(null);
    }
    
    public void RunGrowth(int interval, int trial) {
	_nTrial = trial;
	try
	    {
		// callback every "interval" days
		for (int i=1; i<=_params.get_maxSimulationYears()*365; i+=interval)
		    _scheduler.PostEvent(new SimulationEvent(null, i, SimulationEvent.EventType.callback));
	    }
	catch( Exception ex ) {
	    System.out.println("Exception caught in PNASDriver.RunGrowth(): " + ex.getMessage());
	    ex.printStackTrace(System.out);
	}
	Scheduler.SimulationTerminationReason eTermReason = _scheduler.RunSimulation(this);
    }

    public int get_progressInterval() {
	return Integer.MAX_VALUE; // we never want callbacks based on number of events
    }

    public void Execute(Object o)
    {
	try
	    {
		double day = _scheduler.GetCurrentTime();
		System.out.println(_nTrial + "," + day + "," + _grid.CountCryptsInState(Crypt.eCryptState.AliveMutant) + "," + getNumBorder());
		if (Math.round(day)%365==1 && day>1.0) {
		    File f = new File("trial-"+_nTrial+"-day-"+Math.round(day)+".ps");
		    SaveImage(f);
		}
	    }
	catch( Exception ex ) {
	    System.out.println("Exception caught in PNASDriver.ISimulationCallback.Execute(): " + ex.getMessage());
	    ex.printStackTrace(System.out);
	}
    }

    // save as postscript
    public void SaveImage(File file) {
	SaveImage(file, false);
    }

    // save as postscript
    // bShowNormal - if true, draw normal crypts
    public void SaveImage(File file, boolean bShowNormal)
    {
	try{
	    PrintWriter o = new PrintWriter(new BufferedWriter(new FileWriter(file)));
	    int scale = 3;
	    int cryptsize = 1;
	    int border = 2;
	    o.println("%!PS");
	    o.println("%%BoundingBox: 0 0 " + (border+scale*_grid.getMaxX()+cryptsize) + " " + (border+scale*_grid.getMaxY()+cryptsize));
	    double psscale = 8*72/(scale*_grid.getMaxX()+border); // fit image width to 8 inches
	    o.println(psscale + " " + psscale + " scale");

	    if (bShowNormal) {
		o.println(".85 setgray");
		for (int i=0; i<_grid.getSize(); i++) {
		    Crypt crypt = _grid.getCrypt(i);
		    if (crypt!=null) {
			double x = crypt.getX();
			double y = crypt.getY();
			Color c = _grid.getCrypt(i).getColor();
			if (c!=Color.black) {
			    if (c==Color.white)
				o.println((border+x*scale) + " " + (border+y*scale) + " " + cryptsize + " 0 360 arc fill");
			}
		    }
		}
	    }
	    o.println("0 setgray");
	    for (int i=0; i<_grid.getSize(); i++) {
		Crypt crypt = _grid.getCrypt(i);
		if (crypt!=null) {
		    double x = crypt.getX();
		    double y = crypt.getY();
		    Color c = _grid.getCrypt(i).getColor();
		    if (c!=Color.black && c!=Color.white)
			o.println((border+x*scale) + " " + (border+y*scale) + " " + cryptsize + " 0 360 arc fill");
		}
	    }

	    o.println("showpage");
	    o.close();
	} catch (Exception ex) {
	    System.err.println("File problem" + ex);
	}
    }

    public int print(Graphics g,PageFormat pf,int pageIndex) {
	if (pageIndex == 0) {
	    Graphics2D g2d = (Graphics2D)g;
	    _gridView = new PGridView(_grid);
	    g2d.translate(pf.getImageableX(), pf.getImageableY());
	    // scale page by 4x so that the rows of crypts line up nicely.
	    _gridView.setSize(new Dimension(new Double(pf.getImageableWidth()).intValue()*4, new Double(pf.getImageableHeight()).intValue()*4));
	    g2d.scale(0.1,0.1);
	    paint(g2d);
	    return Printable.PAGE_EXISTS;
	} else {
	    return Printable.NO_SUCH_PAGE;
	}
    }

    public void paint(Graphics g) {
	try
	    {
		_gridView.paintComponent( g );
	    }
	catch( Exception ex )
	    {
		System.err.println( "Exception caught in PNASDriver:Paint(): " + ex.getMessage());
		ex.printStackTrace(System.err);

	    }
    }

    public int getNumDead() {
	return _grid.CountCryptsInState(Crypt.eCryptState.Dead);
    }

    public int getNumMutant() {
	return _grid.CountCryptsInState(Crypt.eCryptState.AliveMutant);
    }

    // number of wildtype or dead crypts that touch a mutant crypt
    public int getNumBorder() {
	int count = 0;
	for (int i=0; i<_grid.getSize(); i++) {
	    Crypt crypt = _grid.getCrypt(i);
	    if (crypt.GetCryptState()!=Crypt.eCryptState.AliveMutant) {
		ArrayList<Integer> neighbors = _grid.getNeighbors(i);
		for (int n=0; n<neighbors.size(); n++) {
		    if (_grid.getCrypt((neighbors.get(n)).intValue()).GetCryptState()==Crypt.eCryptState.AliveMutant) {
			count++;
			break;
		    }
		}
	    }
	}
	return count;
    }

    // number of mutant crypts that touch each border cell
    public int[] getBorderBorders() {
	int count = 0;
	int result[] = new int[7];
	for (int i=0; i<_grid.getSize(); i++) {
	    Crypt crypt = _grid.getCrypt(i);
	    if (crypt.GetCryptState()!=Crypt.eCryptState.AliveMutant) {
		ArrayList<Integer> neighbors = _grid.getNeighbors(i);
		for (int n=0; n<neighbors.size(); n++) {
		    if (_grid.getCrypt((neighbors.get(n)).intValue()).GetCryptState()==Crypt.eCryptState.AliveMutant) { // this is a border cell
			count++;
			int mcount=0;
			ArrayList<Integer> borderneighbors = _grid.getNeighbors(i);
			for (int m=0; m<borderneighbors.size(); m++) {
			    if (_grid.getCrypt((borderneighbors.get(m)).intValue()).GetCryptState()==Crypt.eCryptState.AliveMutant)// this is a mutant
				mcount++;
			}
			result[mcount]++;
			break;
		    }
		}
	    }
	}
	return result;
    }

    /*
     * main - run the driver
     */
    public static void main(String[] args) {
	int experiment = -1;
	try {
	    experiment = Integer.parseInt(args[0]);
	    System.err.println("Running experiment " + experiment);
	}
	catch( Exception ex ) {
	    System.err.println( "Usage: java -server ui.PNASDriver n");
	    System.err.println( "       where n is an experiment number between 1 and 13");
	    System.exit(-1);
	}
	TissueEnvironmentSettings params = new TissueEnvironmentSettings();
	params.set_label("experiment " + experiment);
	KnuthRandom r = new KnuthRandom(); // only used to pick initiation time for some experiments
	r.seedRandom(-1);

	switch (experiment) {
	case 1: // experiment 1: survival advantage growth
	case 2: // experiment 2: reproductive + survival advantage
	case 3: // experiment 3: reproductive advantage with no survival advantage
	case 4: // experiment 4: no reproductive or survival advantage
	    int nNumRuns = 10;
	    int nReportInterval = 365; // output results every n days
	    params.set_gridSizeX(400);
	    params.set_gridSizeY(400);
	    params.set_meanCryptDivisionTime(1.0);
	    params.set_mutationRate(0.0);
	    params.set_wtBackgroundDeathRate(0.01);
	    params.set_mutantWoundResistanceFactor(0.0);// no wounding
	    params.set_wtWoundImpactProbability(0.0);
	    params.set_woundIntervalTime(0);
	    params.set_woundingMode(TissueEnvironmentSettings.WoundingMode.None);
	    if (experiment==1) {
		System.err.println("mutant with survival advantage");
		params.set_mutantReproductiveAdvantageFactor(1.0);
		params.set_mutantSurvivalAdvantageFactor(0.0);     // no mutant death
		params.set_maxSimulationYears(65);
	    } else if  (experiment==2) {
		System.err.println("mutant with reproductive and survival advantage");
		params.set_mutantReproductiveAdvantageFactor(0.1); // reproductive
		params.set_mutantSurvivalAdvantageFactor(0.0);     // no mutant death
		params.set_maxSimulationYears(23);
	    } else if  (experiment==3) {
		System.err.println("mutant with reproductive and no survival advantage");
		params.set_mutantReproductiveAdvantageFactor(0.1); // reproductive
		params.set_mutantSurvivalAdvantageFactor(1.0);     // mutant death
		params.set_maxSimulationYears(23);
	    } else {
		System.err.println("neutral mutant");
		params.set_mutantReproductiveAdvantageFactor(1.0); // no advantage
		params.set_mutantSurvivalAdvantageFactor(1.0);     // mutant death
		params.set_maxSimulationYears(5);
		params.set_gridSizeX(200);                         // smaller grid
		params.set_gridSizeY(200);
		nNumRuns = 100;                                    // more runs
		nReportInterval = 50;                              // more frequent output
	    }
	    // perimeter growth
	    System.out.println("trial,time,mutant,border");
	    for (int i=1; i<=nNumRuns; i++) {
		PNASDriver d = new PNASDriver(params);
		System.err.println("random seed = " +  d.getRandomSeed());
		d.RunGrowth(nReportInterval,i);
		File f = new File("experiment-"+experiment+"-trial-"+i+".ps");
		d.SaveImage(f);
	    }
	    break;
	case 5: // experiment 5: survival advantage - border shape measurements for 1
	case 6: // experiment 6: reproductive advantage - border shape measurements for 2
	    params.set_gridSizeX(200);
	    params.set_gridSizeY(200);
	    params.set_meanCryptDivisionTime(1.0);
	    params.set_mutationRate(0.0);
	    params.set_wtBackgroundDeathRate(0.01);
	    params.set_mutantWoundResistanceFactor(0.0);// no wounding
	    params.set_wtWoundImpactProbability(0.0);
	    params.set_woundIntervalTime(0);
	    params.set_woundingMode(TissueEnvironmentSettings.WoundingMode.None);
	    params.set_mutantSurvivalAdvantageFactor(0.0);  // no mutant death
	    if (experiment==5) {
		System.err.println("mutant with survival advantage");
		params.set_mutantReproductiveAdvantageFactor(1.0);
		params.set_maxSimulationYears(10);
	    } else {
		System.err.println("mutant with reproductive and survival advantage");
		params.set_mutantReproductiveAdvantageFactor(0.1);
		params.set_maxSimulationYears(25);
	    }
	    for (int i=1; i<=10; i++) {
		PNASDriver d = new PNASDriver(params);
		d.Run();
		int c[] = d.getBorderBorders();
		for (int j=1; j<=6; j++) {
		    System.out.print(c[j]);
		    if (j<6)
			System.out.print(",");
		}
		System.out.println();
	    }
	    break;
	case 7: // Brash experiment: immortal p53 mutants, continuous initiation, realistic time scales, lambda supplied by user
	case 8: // Brash experiment: 40\% p53 mutant death, continuous initiation, realistic time scales, lambda supplied by user
	case 9: // Brash experiment: 14\% p53 mutant death, continuous initiation, realistic time scales, lambda supplied by user
	    double lambda = -1.0;
	    if (experiment==7) {
		params.set_mutantSurvivalAdvantageFactor(0.0);  // no mutant death
	    } else if (experiment==8) {
		params.set_mutantSurvivalAdvantageFactor(0.4);  // some mutant death
	    } else if (experiment==9) {
		params.set_mutantSurvivalAdvantageFactor(0.14);  // some mutant death
	    }
	    try {
		lambda = Double.parseDouble(args[1]);
		System.err.println("lambda is  " + lambda);
	    }
	    catch( Exception ex ) {
		System.err.println("lambda must be supplied for experiment number " + experiment + " as the second command line argument");
		System.exit(-1);
	    }
	    if (lambda<0.0) {
		System.err.println("invalid value for lambda " + lambda);
		System.exit(-1);
	    }
	    params.set_wtBackgroundDeathRate(lambda);
	    params.set_gridSizeX(100);
	    params.set_gridSizeY(100);
	    params.set_meanCryptDivisionTime(1.0); // note change
	    params.set_mutantReproductiveAdvantageFactor(1.0);
	    params.set_mutationRate(0.0);
	    params.set_mutantWoundResistanceFactor(0.0);// no wounding
	    params.set_wtWoundImpactProbability(0.0);
	    params.set_woundIntervalTime(0);
	    params.set_woundingMode(TissueEnvironmentSettings.WoundingMode.None);
	    params.set_maxSimulationYears(1);
	    System.out.println("trial,time,mutant,border");
	    for (int i=1; i<=5000; i++) {
		if (i%100==0)
		    System.err.println("Trial "+i);
		PNASDriver d = new PNASDriver(params, 365*params.get_maxSimulationYears()*r.randomDouble());
		System.err.println("random seed = " +  d.getRandomSeed());
		d.RunGrowth(7,i);
	    }
	    break;
	case 10:
	    System.err.println("Experiment 10 is not defined.  Sorry.");
	    break;
	case 11: // experiment 11: Brash experiment: immortal p53 mutants, continuous initiation, realistic time scales, lambda=0.02
	case 12: // experiment 12: lambda=0.04
	case 13: // experiment 13: lambda=0.06
	case 14: // experiment 14: lambda=0.08
	case 15: // experiment 15: lambda=0.10
	case 16: // experiment 16: lambda=0.12
	case 17: // experiment 17: Brash experiment: 40% p53 mutant death, continuous initiation, realistic time scales, lambda=0.02
	case 18: // experiment 18: lambda=0.04
	case 19: // experiment 19: lambda=0.06
	case 20: // experiment 20: lambda=0.08
	case 21: // experiment 21: lambda=0.10
	case 22: // experiment 22: lambda=0.12
	case 23: // experiment 23: lambda=0.14
	    if (experiment<17) {
		params.set_mutantSurvivalAdvantageFactor(0.0);  // no mutant death
		params.set_wtBackgroundDeathRate((experiment-10)*0.02);
	    } else {
		params.set_mutantSurvivalAdvantageFactor(0.4);  // some mutant death
		params.set_wtBackgroundDeathRate((experiment-16)*0.02);
	    }
	    params.set_gridSizeX(100);
	    params.set_gridSizeY(100);
	    params.set_meanCryptDivisionTime(1.0); // note change
	    params.set_mutantReproductiveAdvantageFactor(1.0);
	    params.set_mutationRate(0.0);
	    params.set_mutantWoundResistanceFactor(0.0);// no wounding
	    params.set_wtWoundImpactProbability(0.0);
	    params.set_woundIntervalTime(0);
	    params.set_woundingMode(TissueEnvironmentSettings.WoundingMode.None);
	    params.set_maxSimulationYears(1);
	    System.out.println("trial,time,mutant,border");
	    for (int i=1; i<=5000; i++) {
		if (i%100==0)
		    System.err.println("Trial "+i);
		PNASDriver d = new PNASDriver(params, 365*params.get_maxSimulationYears()*r.randomDouble());
		System.err.println("random seed = " +  d.getRandomSeed());
		d.RunGrowth(7,i);
	    }
	    break;
	} // switch
    }
}
